require 'rest-client'
require 'json'

module Trader
  class BitstampBackend < BaseBackend
    BASE_CUR = Currency.for_code(:BTC)
    QUOTE_CUR = Currency.for_code(:USD)
    MAIN_MARKET = CurrencyPair.new BASE_CUR, QUOTE_CUR

    TYPE_MAP = {
      0 => Order::BID,
      1 => Order::ASK
    }

    class BackendBalance < Struct.new(:amount, :available_amount)
      # nothing for now
    end

    class BitstampOrder < RawAccountOrder
      attr_mapped(:id, 'raw') # use original order information as id
      attr_mapped(:pair) { |r| MAIN_MARKET }
      attr_mapped(:price) { |r| r['raw']['price'] }
      attr_mapped(:volume) { |r| r['raw']['amount'] }
      attr_mapped(:executed_volume)
      attr_mapped(:instruction) { |r| r['raw']['type'] == 0 ? Order::BID : Order::ASK }

      attr_mapped(:status) do |r|
        case r['status']
        when 'Open'
          AccountOrder::OPEN
        when 'Finished'
          r['executed_volume'] < big(r['raw']['amount']) ? AccountOrder::CANCELED : AccountOrder::CLOSED
        else
          AccountOrder::PENDING
        end
      end

      def limit?
        true
      end
    end

    def initialize
      super :bitstamp
    end

    def get_available_markets
      [MAIN_MARKET]
    end

    def fill_book(_book)
      # TODO: consider book pair

      _book.prepare Time.now

      ob = execute_request(nil, 'order_book')
      ob['bids'].each { |o| _book.add_bid(o[0], o[1]) }
      ob['asks'].each { |o| _book.add_ask(o[0], o[1]) }

      tx = execute_request(nil, 'transactions')
      tx.each do |t|
        _book.add_transaction t['price'], t['amount'], Time.at(t['date'].to_i)
      end
    end

    def get_balance(_session, _currency)
      raw = execute_request(_session || session, 'balance')

      if _currency == BASE_CUR
        return BackendBalance.new raw['btc_balance'], raw['btc_available']
      else
        return BackendBalance.new raw['usd_balance'], raw['usd_available']
      end
    end

    def get_orders(_session, _pair)
      raw_orders = execute_request(_session || session, 'open_orders')
      raw_orders.map { |o| normalize_raw_order o }
    end

    def create_order(_session, _pair, _volume, _price, _type)
      not_supported('market orders') if _price.nil?

      normalize_raw_order execute_request(
        _session || session,
        _type == Order::BID ? 'buy' : 'sell',
        { amount: _volume.to_f, price: _price.to_f }
      )
    end

    def fetch_order(_session, _id)
      normalize_raw_order_status _id, execute_request(
        _session || session,
        'order_status',
        { id: _id['id'] }
      )
    end

    private

    attr_reader :session

    def execute_request(_signing, _resource, _data=nil)
      _data = sign_params _signing, _data if _signing

      JSON.parse(if _data.nil?
        RestClient.get "https://www.bitstamp.net/api/#{_resource}/"
      else
        RestClient.post "https://www.bitstamp.net/api/#{_resource}/", _data
      end)
    end

    def sign_params(_keys, _params)
      nonce = (Time.now.to_f*10000).to_i.to_s

      customer_id = _keys[:customer_id].to_s
      api_key = _keys[:api_key]
      api_secret = _keys[:api_secret]

      digest = OpenSSL::Digest.new('sha256')
      signature_data = "#{nonce}#{customer_id}#{api_key}"
      signature = OpenSSL::HMAC.hexdigest(digest, api_secret, signature_data)

      (_params || {}).merge({
        nonce: nonce,
        signature: signature.upcase,
        key: api_key
      })
    end

    def normalize_raw_order(_order)
      BitstampOrder.new({
        'raw' => _order,
        'executed_volume' => big(0),
        'status' => 'In Queue'
      })
    end

    def normalize_raw_order_status(_id, _status)
      executed_volume = _status['transactions'].inject(big(0)) { |r, t| r + big(t['btc']) }
      executed_price = _status['transactions'].inject(big(0)) { |r, t| r + (big(t['price']) * big(t['btc'])) }

      BitstampOrder.new({
        'raw' => _id,
        'executed_volume' => executed_volume,
        'executed_price' => executed_price / executed_volume,
        'status' => _status['status']
      })
    end

    def big(_string)
      BigDecimal.new _string
    end
  end
end